import time


def printlog(**kwargs):
    ctime = time.strftime('%Y-%m-%d/%H:%M:%S', time.localtime())
    msg = f'ctime={ctime}'
    for k in kwargs:
        msg += f'\t| {k}={kwargs[k]}'
    print(msg)



from torch.utils.data import Dataset, DataLoader


def is_enum(obj):
    return type(obj) is tuple or type(obj) is list


def datasets(dataset_class, transforms=(None, None, None), **kwargs):
    datasets_collection = [
        dataset_class(train=True, transform=transforms[0], **kwargs),
        dataset_class(train=False, transform=transforms[1], **kwargs),
        dataset_class(train=False, transform=transforms[2], **kwargs),
    ]
    # check transforms
    assert len(transforms) == 3
    for d in datasets_collection:
        try:
            _ = d.__getitem__(0)
        except Exception as e:
            raise e
    return datasets_collection


def dataloaders(datasets,  **kwargs):
    # check datasets input
    assert len(datasets) == 3
    for x in datasets:
        assert issubclass(type(x), Dataset)
        try:
            x.__getitem__(0)
        except Exception as e:
            raise e
    trainset, testset, valset = datasets
    trainloader = DataLoader(trainset, shuffle=True, **kwargs)
    testloader = DataLoader(testset, **kwargs)
    valloader = DataLoader(valset, **kwargs)
    return trainloader, testloader, valloader



import os

from torchmods.utils import bz2

from . import ssh


class _DeviceInterface():
    """ Communication with device via ssh tunnel.
    """

    def __init__(self, key):
        self.__key = key

    def exec_command(self, root, cmd):
        with ssh.open(self.__key) as f:
            out = f.cd(root).call(cmd)
        return out

    def push(self, localpath, remotepath):
        with ssh.open(self.__key) as f:
            print(localpath, remotepath)
            f.upload(localpath, remotepath)
        return self

    def pull(self, remotepath, localpath):
        with ssh.open(self.__key) as f:
            f.download(remotepath, localpath)
        return self


class _DeviceEnvironmentBlock(_DeviceInterface):
    """ A block of occupied device with full I/O access.
    """

    def __init__(self, key, root):
        super().__init__(key)
        self.__root = root
        self.__key = key
        self.create_dir(root)

    def create_dir(self, dir):
        with ssh.open(self.__key) as f:
            f.call(f'[ -d {dir} ] || mkdir {dir}')
        return self

    def exec_command(self, cmd):
        return super().exec_command(self.__root, cmd)

    def cat1stline(self, filename):
        out = self.exec_command('cat ' + filename)
        out = out[0].replace('\n', '')
        return out

    def clear(self):
        super().exec_command('.', f'rm -r {self.__root}/*')
        return self

    def push(self, localpath):
        filename = localpath.split('/')[-1]
        super().push(localpath, self.__root + '/' + filename)
        return self

    def pull(self, filepath, localroot):
        filename = filepath.split('/')[-1]
        super().pull(self.__root + '/' + filepath, localroot + '/' + filename)
        return self

    def pull_all(self, localroot):
        for filename in self.exec_command('ls'):
            filename = filename.replace('\n', '')
            self.pull(filename, localroot)
        return self

    def path(self):
        # used for generating local mirror dir,
        # therefore we check if local path exists first.
        if not os.path.exists(self.__root):
            os.makedirs(self.__root)
        return self.__root


class _IOCache():
    """ Feed source files,
        ** Device, do yours, **
        Get packed outputs.
    """

    def __init__(self, key, root):
        self.__input = _DeviceEnvironmentBlock(key, root + '/input')
        self.__output = _DeviceEnvironmentBlock(key, root + '/output')

    def push(self, tar):
        self.__input.push(tar)
        return self

    def pull(self):
        mirror_dir = self.__output.path()
        self.__output.pull_all(mirror_dir)
        return self


class _BackServer():
    """ A handler to wake/control the offline server. (which load tasks and run infinitely)
    """

    def __init__(self, key, root, server_script=None):
        self.__key = key
        self.__root = root + '/server'
        self.__core = _DeviceEnvironmentBlock(self.__key,  self.__root)
        self.__vm = _DeviceEnvironmentBlock(self.__key, root + '/vm')
        self.server_script = server_script

    def awake(self):
        with ssh.open(self.__key) as f:
            f.call(f'[ -f {self.__root}/awake ] || echo False>{self.__root}/awake')
        flag = (self.__core.cat1stline('awake') == 'True')
        return flag

    def wake(self):
        if self.awake():
            return
        if self.server_script is None:
            path = self.default_server_script()
        self.__core.push(path)
        self.__core.exec_command('chmod +x ./server.sh;nohup ./server.sh>>history 2>&1 & echo $!>server_pid')

    def default_server_script(self):
        return os.path.dirname(os.path.abspath(__file__)) + '/server.sh'


class Device(_DeviceEnvironmentBlock):
    """ Consider remote CUDA server as just DEVICE.
        put task, get processed.
        maybe check the status?
        we dont need anything else actually.
    """

    def __init__(self, key, root):
        super().__init__(key, root)
        self.io = _IOCache(key, root)
        self.core = _BackServer(key, root)

    def build(self, tid, source, ignore):
        tar_name = bz2.compress(source, self.path(), tid, ignore)
        tar = f'{self.path()}/{tar_name}.tar.bz2'
        return tar

    def put(self, tid, source, ignore=None):
        tar = self.build(tid, source, ignore=None)
        self.io.push(tar)
        self.core.wake()
        os.system('rm ' + tar)
        return self

    def get(self):
        self.io.pull()
        return self

    def check_something(self):
        # if we really need this.
        return


def mount(key, root):
    return Device(key, root)




import torch
from torch import nn as nn


class Accuracy(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        if reduction != 'batchmean':
            raise NotImplementedError

    def forward(self, input, target):
        if len(target.shape) > 1:
            target = target.argmax(dim=1)
        _, input = input.max(dim=1)
        acc = input.eq(target).sum().float()/target.size(0)
        return acc


class CorrelationCoefficientsLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        if reduction != 'batchmean':
            raise NotImplementedError

    def forward(self, input, target):
        batch_size = input.shape[0]
        input = input.view(input.size(0), -1)
        target = target.view(target.size(0), -1)
        CC = []
        for i in range(batch_size):
            im = input[i] - torch.mean(input[i])
            tm = target[i] - torch.mean(target[i])
            CC.append(torch.sum(im * tm) / (torch.sqrt(torch.sum(im ** 2))
                                            * torch.sqrt(torch.sum(tm ** 2))))
            CC[i].unsqueeze_(0)
        CC = torch.cat(CC, 0)
        CC = torch.mean(CC)
        return CC


class ChebyshevLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        if reduction != 'batchmean':
            raise NotImplementedError

    def forward(self, input, target):
        cheb = (input-target).abs().max(dim=1)[0]
        cheb = cheb.mean()
        return cheb


class ClarkLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        if reduction != 'batchmean':
            raise NotImplementedError

    def forward(self, input, target):
        clark = (input-target)/(input+target)
        clark = torch.norm(clark, dim=1).mean()
        return clark


class CanberLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        if reduction != 'batchmean':
            raise NotImplementedError

    def forward(self, input, target):
        canber = (input-target).abs()/(input+target)
        canber = canber.sum(dim=1).mean()
        return canber


class CosineLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        if reduction != 'batchmean':
            raise NotImplementedError

    def forward(self, input, target):
        cosine = nn.functional.cosine_similarity(input, target, dim=1).mean()
        return cosine


class IntersecLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        if reduction != 'batchmean':
            raise NotImplementedError

    def forward(self, input, target):
        intersec = input.min(target).sum(dim=1).mean()
        return intersec


class SoftmaxWrapper(nn.Module):
    def __init__(self, core, softmax_input=True, softmax_target=False):
        super().__init__()
        self.core = core
        self.softmax_input = softmax_input
        self.softmax_target = softmax_target

    def adaptive_softmax(self, x):
        if len(x.shape) > 2:
            # some issue with torch.nn.softmax2d so we use this.
            shape = x.shape
            x = x.flatten(start_dim=2)
            x = x.softmax(dim=2)
            x = x.view(shape)
        else:
            x = x.softmax(dim=1)
        return x

    def forward(self, src, target):
        if self.softmax_input:
            src = self.adaptive_softmax(src)
        if self.softmax_target:
            target = self.adaptive_softmax(target)
        return self.core(src, target)


class LogSoftmaxWrapper(nn.Module):
    def __init__(self, core):
        super().__init__()
        self.core = core

    def adaptive_softmax(self, x):
        if len(x.shape) > 2:
            # some issue with torch.nn.softmax2d so we use this.
            shape = x.shape
            x = x.flatten(start_dim=2)
            x = x.log_softmax(dim=2)
            x = x.view(shape)
        else:
            x = x.log_softmax(dim=1)
        return x

    def forward(self, src, target):
        src = self.adaptive_softmax(src)
        return self.core(src, target)


class LogWrapper(nn.Module):
    def __init__(self, core):
        super().__init__()
        self.core = core

    def forward(self, src, target):
        return self.core(torch.log(src), target)


class Softmax2d_(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        shape = x.shape
        x = x.flatten(start_dim=2)
        x = x.softmax(dim=2)
        x = x.view(shape)
        return x


if __name__ == "__main__":
    loss = SoftmaxWrapper(CorrelationCoefficientsLoss()).cuda()
    x = torch.rand(4, 1, 30, 40).cuda()
    target = torch.rand(4, 1, 30, 40).cuda()
    print(loss(x, target))



import json
import os
import random

import numpy as np
import torch
import torch.utils.data as data
import torchmods.nn as nn
import torchvision.transforms as transforms
from PIL import Image
from torchmods.nn.atomic import inference, step
from torchmods.utils import find_path, printlog

from .dataset_collection import dataloaders, datasets


def __maxidx__(_list):
    return _list.index(max(_list))


class LDL(data.Dataset):
    img_dir = 'images'
    seed = 0

    def __init__(self, root, train=True, transform=None, lb_type='dist'):
        super().__init__()
        self.root = root
        self.transform = transform
        self.lb_type = lb_type
        self.load_data(root, train)
        assert lb_type in ['dist', 'onehot']
        if lb_type == 'dist':
            self.dist()
        elif lb_type == 'onehot':
            self.onehot()

    def __gensplit__(self, root):
        """
            generate train/test split as split.json,
            Args:
                root:
        """
        np.random.seed(self.seed)
        _list = [x for x in os.listdir(f'{root}/{self.img_dir}')]
        random.shuffle(_list)
        pivod = int(len(_list)*0.2)
        _dict = {'train': _list[pivod:], 'test': _list[:pivod]}
        with open(f'{root}/split.json', 'w') as f:
            json.dump(_dict, f)

    def __readgt__(self, root, spliter=' ', start_line=1, suffix=''):
        """
            read groundtruth from ground_truth.txt,
            (override if more complex func needed),
            Args:
                root:
                spliter: spliter for labels
                start_line: line to start from
                suffix: image file suffix
        """
        _dict = {}
        with open(f'{root}/ground_truth.txt', 'r') as f:
            for line in f.readlines()[start_line:]:  # images-name Amusement Awe Contentment Excitement Anger Disgust Fear Sadness
                data = line.split(spliter)
                _dict[f'{data[0]}{suffix}'] = data
        return [_dict[x.split('/')[-1]] for x in self.data]

    # def __cvimg__(self, path, convert2rgb=True):
    #     img_cv = cv2.imread(path)
    #     if convert2rgb:
    #         img_cv = cv2.cvtColor(img_cv, cv2.COLOR_BGR2RGB)
    #     return img_cv

    def __getitem__(self, index):
        # deprecated method (because of thread lock issue:https://zhuanlan.zhihu.com/p/133707658)
        # img = self.__cvimg__(self.data[index])
        img = Image.open(self.data[index]).convert('RGB')
        if self.transform:
            img = self.transform(img)
        target = self.labels[index]
        return img, target

    def __len__(self):
        return len(self.data)

    def load_data(self, root, train):
        assert os.path.exists(root)
        _path = f'{root}/split.json'
        if not os.path.exists(_path):
            self.__gensplit__(root)
        with open(_path, 'r') as f:
            _json = json.load(f)
            self.data = [f'{root}/{self.img_dir}/{x}' for x in _json['train' if train else 'test']]

    def load_label(self, root):
        for x in self.__readgt__(root):
            y = [int(_str.replace('\n', '')) for _str in x[1:]]
            if self.lb_type == 'onehot':
                y = __maxidx__(y)
                y = torch.tensor(y)
            elif self.lb_type == 'dist':
                y = (np.array(y)/sum(y)).astype(np.float32)
                y = torch.from_numpy(y)
            self.labels.append(y)

    def dist(self):
        self.lb_type == 'dist'
        self.labels = []
        self.load_label(self.root)

    def onehot(self):
        self.lb_type == 'onehot'
        self.labels = []
        self.load_label(self.root)


class EnvLDL():
    def __init__(self):
        self.root = find_path(['/home/sh/datasets/Flickr_LDL', '/home/public/datasets/Flickr_LDL'])
        self.transforms = (
            transforms.Compose([
                transforms.Resize(size=(240, 320)),
                transforms.RandomHorizontalFlip(),
                transforms.RandomRotation(degrees=15),
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
            ]),
            transforms.Compose([
                transforms.Resize(size=(240, 320)),
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
            ]),
            transforms.Compose([
                transforms.Resize(size=(240, 320)),
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
            ])
        )
        self.max_training_patience = 8
        self.checkpoints = torch.linspace(0, 1, steps=20)
        self.max_epoch = 1000
        self.batch_size = 32
        self.num_workers = 2
        self.pin_memory = False

    def sample(self, batch_size=1, shuffle=True):
        from torch.utils.data import DataLoader
        for batch in DataLoader(LDL(root=self.root, train=False, transform=self.transforms[-1]), batch_size=batch_size, shuffle=shuffle):
            return batch

    def check(self, model, optimizer):
        torch.autograd.set_detect_anomaly(True)
        self.train(model, optimizer, _test_mode=True)
        self.performance(model, _test_mode=True)
        print('> passed.')

    def train(self, model, optimizer, _test_mode=False):
        trainloader, valloader, testloader = dataloaders(
            datasets(LDL, root=self.root, transforms=self.transforms),
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )
        min_average_kldiv_test = float('inf')
        patience = self.max_training_patience
        criterion = nn.LogSoftmaxWrapper(nn.KLDivLoss(reduction='batchmean'))
        checkpoints = (self.checkpoints*len(trainloader)).numpy().astype('int').tolist()
        printlog(checkpoints=checkpoints)
        for epoch in range(self.max_epoch):
            # train model on full trainset
            model.train()
            loss_patch = []
            for iteration, (x, target) in enumerate(trainloader):
                x, target = x.cuda(), target.cuda()
                loss_patch.append(step(model, x, target, criterion, optimizer))
                if iteration in checkpoints:
                    loss = sum(loss_patch)/len(loss_patch)
                    printlog(epoch=epoch, iter=iteration, loss=loss, mode='train')
                    loss_patch = []
                if _test_mode:
                    break
            # validation on full valset
            model.eval()
            loss_patch = []
            with torch.no_grad():
                for iteration, (x, target) in enumerate(valloader):
                    x, target = x.cuda(), target.cuda()
                    loss_patch.append(inference(model, x, target, criterion))
                    if _test_mode:
                        break
            loss = sum(loss_patch)/len(loss_patch)
            printlog(epoch=epoch, iter=iteration, loss=loss, mode='validation')
            if _test_mode:
                break
            if loss < min_average_kldiv_test:
                min_average_kldiv_test = loss
                patience = self.max_training_patience
                torch.save(model.state_dict(), f'last.pth')
            else:
                patience -= 1
                if patience < 1:
                    break
        return model

    def performance(self, model,  verbose=False, _test_mode=False):
        testloader = torch.utils.data.DataLoader(
            LDL(root=self.root, train=False, transform=self.transforms[-1]),
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )
        # test performance
        model.eval()
        softmax = nn.Softmax(dim=1)
        metrics_patch = {
            'Acc': [nn.Accuracy(), []],
            'Cheb': [nn.ChebyshevLoss(), []],
            'Clark': [nn.ClarkLoss(), []],
            'Canber': [nn.CanberLoss(), []],
            'KLDiv': [nn.LogWrapper(nn.KLDivLoss(reduction='batchmean')), []],
            'Cosine': [nn.CosineLoss(), []],
            'Intersec': [nn.IntersecLoss(), []],
        }
        with torch.no_grad():
            for iteration, (x, target) in enumerate(testloader):
                x, target = x.cuda(), target.cuda()
                y = softmax(model(x))
                logs = {'iter': iteration, 'mode': 'test'}
                for m in metrics_patch:
                    metrics_patch[m][1].append(metrics_patch[m][0](y, target).item())
                    v = sum(metrics_patch[m][1])/len(metrics_patch[m][1])
                    logs[m] = v
                if verbose:
                    printlog(**logs)
                if _test_mode:
                    break
            printlog(**logs)
        return metrics_patch





import paramiko
import base64


class Tunnel():
    def __init__(self, key):
        self.__key = key
        self.__client = None
        self.__root = '~'

    @staticmethod
    def decode(key: str):
        sequence = base64.b64decode(key)
        return str(sequence, 'utf8').split('$')

    @staticmethod
    def connect(key):
        ssh = paramiko.SSHClient()  # 创建sshclient
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())  # 指定当对方主机没有本机公钥的情况时应该怎么办，AutoAddPolicy表示自动在对方主机保存下本机的秘钥
        ssh.connect(*Tunnel.decode(key))
        return ssh

    @staticmethod
    def sftp_callback(transferred, total):
        print(f'\rtransferred/total: {transferred}/{total} ({transferred/total:%})', end='')

    def __enter__(self):
        self.__client = self.connect(self.__key)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__client.close()

    def __call__(self, cmd):
        _, stdout, stderr = self.__client.exec_command(f'. /etc/profile;. .bashrc;cd {self.__root};{cmd}')
        out = stdout.readlines()
        err = stderr.readlines()
        if len(err) > 0:
            print(err[0])
            raise err[0]
        return out

    def upload(self, local, remote):
        with self.__client.open_sftp() as sftp:
            sftp.put(local, remote, callback=self.sftp_callback)
        print()
        return self

    def download(self, remote, local):
        with self.__client.open_sftp() as sftp:
            sftp.get(remote, local, callback=self.sftp_callback)
        print()
        return self

    def cd(self, arg1):
        self.call('cd ' + arg1)
        self.__root = arg1
        self.__root = self.call('pwd')[0].replace('\n', '')
        return self

    def call(self, command):
        return self.__call__(command)


def keygen(ip_address, user, password, port=22):
    sequence = f'{ip_address}${port}${user}${password}'
    return str(base64.b64encode(sequence.encode('utf8')), 'utf8')


def open(key):
    return Tunnel(key)
